import math
from typing import Optional, Any

from fmi_weather_client.models import FMIObservation


def float_or_none(v: Any) -> Optional[float]:
    """
    Get value as float. None if conversion fails.
    :param v: Any value
    :return: Value as float if successful; None otherwise
    """
    try:
        f = float(v)
        if not math.isnan(f):
            return f
        return None
    except (ValueError, TypeError):
        return None


def is_non_empty_observation(observation: FMIObservation) -> bool:
    """
    Validate if observation contains proper values
    :param observation: Observation
    :return: True if observation is not empty; False otherwise
    """
    for _, value in observation.variables.items():
        if value is not None:
            return True
    return False



import os


class MockElapsed:
    def __init__(self):
        self.microseconds = 10000000


class MockResponse:

    def __init__(self, xml, status_code):
        self.text = xml
        self.status_code = status_code
        self.elapsed = MockElapsed()


def mock_place_response(*args, **kwargs):
    return __mock_response('valid_place_response.xml', 200, args, kwargs)


def mock_bbox_response(*args, **kwargs):
    return __mock_response('valid_bbox_response.xml', 200, args, kwargs)


def mock_empty_response(*args, **kwargs):
    return __mock_response('empty_response.xml', 200, args, kwargs)


def mock_nan_response(*args, **kwargs):
    return __mock_response('nan_response.xml', 200, args, kwargs)


def mock_no_location_exception_response(*args, **kwargs):
    return __mock_response('exception_no_locations_response.xml', 200, args, kwargs)


def mock_other_exception_response(*args, **kwargs):
    return __mock_response('exception_other_response.xml', 200, args, kwargs)


def __mock_response(filename, status_code, *args, **kwargs):
    dirname = os.path.dirname(__file__)
    xml_file = os.path.join(dirname, filename)
    with open(xml_file, 'r') as mock_file:
        xml = mock_file.read()

    return MockResponse(xml, status_code)



import logging
from datetime import datetime, timedelta
from typing import Any, Dict

import requests

from fmi_weather_client import parser
from fmi_weather_client.errors import ServiceError
from fmi_weather_client.models import Weather

_BASE_URL = 'http://opendata.fmi.fi/wfs'

_DEFAULT_PARAMS = {
    'service': 'WFS',
    'version': '2.0.0',
    'request': 'getFeature',
    'storedquery_id': 'fmi::observations::weather::multipointcoverage',
    'timestep': '10'
}

_LOGGER = logging.getLogger(__name__)


def weather_by_coordinates(lat: float, lon: float) -> Weather:
    """
    Get the latest weather information by coordinates.

    180 km x 180 km bounding box used to search the closest weather station. Observations are from the past hour.

    :param lat: Latitude (e.g. 25.67087)
    :param lon: Longitude (e.g. 62.39758)
    :return: Latest weather information from the closest weather station
    """
    params = {
        'bbox': '%s,%s,%s,%s' % (lon - 0.8, lat - 0.8, lon + 0.8, lat + 0.8),
        'starttime': (datetime.utcnow() + timedelta(hours=-1)).isoformat(timespec='seconds')
    }
    response = _request_fmi(params)
    return parser.parse_weather_data(response, lat, lon)


def weather_by_place_name(name: str) -> Weather:
    """
    Get the latest weather information by place name.

    Search relies of FMI's own service. Observations are from the past hour.

    :param name: Place name (e.g. Kaisaniemi,Helsinki)
    :return: Latest weather information from the closest weather station
    """
    params = {
        'place': name.strip().replace(' ', ''),
        'starttime': (datetime.utcnow() + timedelta(hours=-1)).isoformat(timespec='seconds')
    }
    response = _request_fmi(params)
    return parser.parse_weather_data(response)


def weather_multi_station(lat: float, lon: float) -> Weather:
    """
    Get the latest full weather by combining data from multiple weather stations
    :param lat: Latitude
    :param lon: Longitude
    :return: Latest weather information from multiple weather stations
    """
    params = {
        'bbox': '%s,%s,%s,%s' % (lon - 0.8, lat - 0.8, lon + 0.8, lat + 0.8),
        'starttime': (datetime.utcnow() + timedelta(hours=-1)).isoformat(timespec='seconds')
    }
    response = _request_fmi(params)
    return parser.parse_multi_weather_data(response, lat, lon)


def _request_fmi(params: Dict[str, Any]) -> str:
    """
    Send a request to FMI service and return the body
    :param params: Query parameters
    :return: Response body
    """

    final_params = _DEFAULT_PARAMS.copy()
    final_params.update(params)

    _LOGGER.debug("Sending GET to %s with parameters: %s", _BASE_URL, final_params)
    response = requests.get(_BASE_URL, params=final_params)

    if response.status_code != 200:
        raise ServiceError("Invalid FMI service response", {'status_code': response.status_code, 'body': response.text})

    _LOGGER.debug("Received a response from FMI in %s ms", response.elapsed.microseconds / 1000)
    return response.text



from datetime import datetime
from typing import Optional, List, Dict


class FMIStation:
    def __init__(self, name: str, lat: float, lon: float):
        self.name = name
        self.lat = lat
        self.lon = lon


class FMIObservation:
    def __init__(self, timestamp: datetime, lat: float, lon: float):
        self.timestamp: datetime = timestamp
        self.lat: float = lat
        self.lon: float = lon
        self.variables: Dict[str, Optional[float]] = {}


class FMIStationObservation:
    def __init__(self, station: FMIStation, observations: List[FMIObservation]):
        self.station = station
        self.observations = observations


class WeatherMeasurement:
    def __init__(self, value: float, unit: str):
        self.value: float = value
        self.unit: str = unit

    def __str__(self):
        return "%s %s" % (self.value, self.unit)


class Weather:
    def __init__(self, station_name: str, station_lat: float, station_lon: float, observation: FMIObservation):

        def try_get_value(o: FMIObservation, variable_name: str, unit: str):
            value = o.variables.get(variable_name, None)
            if value is not None:
                return WeatherMeasurement(value, unit)
            return None

        self.station_name: str = station_name
        self.station_lat: float = station_lat
        self.station_lon: float = station_lon
        self.observation_time: datetime = observation.timestamp

        self.temperature: Optional[WeatherMeasurement] = try_get_value(observation, 't2m', '°C')
        self.humidity: Optional[WeatherMeasurement] = try_get_value(observation, 'rh', '%')
        self.wind_speed: Optional[WeatherMeasurement] = try_get_value(observation, 'ws_10min', 'm/s')
        self.wind_gust: Optional[WeatherMeasurement] = try_get_value(observation, 'wg_10min', 'm/s')
        self.wind_direction: Optional[WeatherMeasurement] = try_get_value(observation, 'wd_10min', '°')
        self.dew_point: Optional[WeatherMeasurement] = try_get_value(observation, 'td', '°C')
        self.precipitation_amount: Optional[WeatherMeasurement] = try_get_value(observation, 'r_1h', 'mm')
        self.precipitation_intensity: Optional[WeatherMeasurement] = try_get_value(observation, 'ri_10min', 'mm/h')
        self.pressure: Optional[WeatherMeasurement] = try_get_value(observation, 'p_sea', 'hPa')
        self.visibility: Optional[WeatherMeasurement] = try_get_value(observation, 'vis', 'm')
        self.cloud_coverage: Optional[WeatherMeasurement] = try_get_value(observation, 'n_man', '1/8')
        self.snow_depth: Optional[WeatherMeasurement] = try_get_value(observation, 'snow_aws', 'cm')
        # Current weather code (In Finnish: https://www.ilmatieteenlaitos.fi/latauspalvelun-pikaohje)
        self.wawa: Optional[WeatherMeasurement] = try_get_value(observation, 'wawa', '')

        # TODO: Support wawa
        # https://helda.helsinki.fi/bitstream/handle/10138/37284/PRO_GRADU_BOOK_HERMAN.pdf?sequence=2



import fmi_weather_client


def print_weather(weather):
    print(weather.station_name)
    print(weather.observation_time)
    print('Temperature: %s' % weather.temperature)
    print('Dew point: %s' % weather.dew_point)
    print('Humidity: %s' % weather.humidity)
    print('Wind speed: %s' % weather.wind_speed)
    print('Wind gusts: %s' % weather.wind_gust)
    print('Wind direction: %s' % weather.wind_direction)
    print('Precipitation intensity: %s' % weather.precipitation_intensity)
    print('Precipitation amount: %s' % weather.precipitation_amount)
    print('Pressure: %s' % weather.pressure)
    print('Visibility: %s' % weather.visibility)
    print('Cloud coverage: %s' % weather.cloud_coverage)
    print('Snow depth: %s' % weather.snow_depth)
    print('WaWa code: %s', weather.wawa)


weather = fmi_weather_client.weather_by_coordinates(60.170998, 24.941325)
weather2 = fmi_weather_client.weather_by_place_name("Kuopio")
weather3 = fmi_weather_client.weather_multi_station(69.016989, 21.465569)

print_weather(weather)
print('')
print_weather(weather2)
print('')
print_weather(weather3)



import math
import sys
from datetime import datetime, timezone
from typing import Dict, Any, List
from typing import Optional

import xmltodict

from fmi_weather_client import errors, utils
from fmi_weather_client.models import FMIStation, FMIObservation, FMIStationObservation, Weather


def parse_weather_data(body: str,
                       lat: Optional[float] = None,
                       lon: Optional[float] = None) -> Weather:
    """
    Parse weather information from FMI response
    :param body: HTTP response body from FMI
    :param lat: latitude
    :param lon: longitude
    :return: Weather information
    """
    data = _body_to_dict(body)
    all_observations = _get_observations_per_station(data)

    # Weather by place name return only one station data so let's use it as a default because why not
    closest_observation_set = all_observations[0]

    if lat is not None and lon is not None:
        closest_observation_set = _get_closest_station_observations(lat, lon, all_observations)

    # It is possible that name search or closest station provides no data. Throw an error in that case.
    if len(closest_observation_set.observations) == 0:
        raise errors.NoWeatherDataError

    latest_observation = closest_observation_set.observations[-1]

    return Weather(closest_observation_set.station.name,
                   closest_observation_set.station.lat,
                   closest_observation_set.station.lon,
                   latest_observation)


def parse_multi_weather_data(body: str,
                             lat: float,
                             lon: float) -> Weather:
    """
    Parse weather information from FMI response
    :param body: HTTP response body from FMI
    :param lat: latitude
    :param lon: longitude
    :return: Weather information
    """
    data = _body_to_dict(body)
    all_observations = _get_observations_per_station(data)
    sorted_observation_set = _get_observations_sorted_by_distance(lat, lon, all_observations)

    closest_station = sorted_observation_set[0].station
    latest_observation = sorted_observation_set[0].observations[-1].timestamp

    observations = FMIObservation(latest_observation, lat, lon)
    for station_observation in sorted_observation_set:
        if len(station_observation.observations) > 0:
            latest_variables = station_observation.observations[-1].variables
            for variable_name, value in latest_variables.items():
                if observations.variables.get(variable_name, None) is None and value is not None:
                    observations.variables[variable_name] = value

    return Weather(closest_station.name,
                   closest_station.lat,
                   closest_station.lon,
                   observations)


def _body_to_dict(body: str) -> Dict[str, Any]:
    """
    Convert FMI response body to dictionary and check errors
    :param body: FMI response body
    :return: Response as dictionary
    """
    data = xmltodict.parse(body)

    # Check exception response
    if 'ExceptionReport' in data.keys():
        if 'No locations found for the place' in data['ExceptionReport']['Exception']['ExceptionText'][0]:
            raise errors.NoWeatherDataError
        raise errors.ServiceError(data['ExceptionReport']['Exception']['ExceptionText'][0])

    if 'wfs:member' not in data['wfs:FeatureCollection'].keys():
        raise errors.NoWeatherDataError

    return data


def _get_stations(data: Dict[str, Any]) -> List[FMIStation]:
    """
    Try to get station data from the response. When call is made with place name, there is only one
    station available. For coordinate search there might be more.
    :param data: Response data from FMI as xmltodict object
    :return: List of stations
    """

    def parse_station_data(station_data: Dict[str, Any]) -> FMIStation:
        name = station_data['gml:Point']['gml:name']
        position = station_data['gml:Point']['gml:pos'].split(' ', 1)
        return FMIStation(name,
                          float(position[0]),
                          float(position[1]))

    stations = []
    stations_dict = (data['wfs:FeatureCollection']['wfs:member']['omso:GridSeriesObservation']
                     ['om:featureOfInterest']['sams:SF_SpatialSamplingFeature']['sams:shape']['gml:MultiPoint']
                     ['gml:pointMember'])

    # xmltodict can return a list or an object depending on how many child the element has
    if isinstance(stations_dict, list):
        for station_dict in stations_dict:
            station = parse_station_data(station_dict)
            stations.append(station)
    else:
        stations.append(parse_station_data(stations_dict))

    return stations


def _get_observation_types(data: Dict[str, Any]) -> List[str]:
    """
    Try to get available observation types (temperature, pressure etc) from the response.
    Available types depends on the available stations.
    :param data: Response data from FMI as xmltodict object
    :return: List of observation types
    """
    observation_types = []
    observation_types_data = (data['wfs:FeatureCollection']['wfs:member']['omso:GridSeriesObservation']['om:result']
                              ['gmlcov:MultiPointCoverage']['gmlcov:rangeType']['swe:DataRecord']['swe:field'])
    for observation_type in observation_types_data:
        observation_types.append(observation_type['@name'])
    return observation_types


def _get_observations(data: Dict[str, Any]) -> List[FMIObservation]:
    observation_types = _get_observation_types(data)

    # Get observation times and values for matching. They are different elements in XML but the amount of
    # data points should be the same so times and values can be matched.
    #
    # Time data format is always the same and fields are separated by space:
    # station_longitude station_latitude unix_timestamp
    #
    # Values are also separated by space but number of fields depends on the different variables stations provide.
    # The number of fields should be the same as number of available observation types so these two can be matched.
    #
    observation_times_list = (data['wfs:FeatureCollection']['wfs:member']['omso:GridSeriesObservation']['om:result']
                              ['gmlcov:MultiPointCoverage']['gml:domainSet']['gmlcov:SimpleMultiPoint']
                              ['gmlcov:positions'].split('\n'))
    observation_values_list = (data['wfs:FeatureCollection']['wfs:member']['omso:GridSeriesObservation']['om:result']
                               ['gmlcov:MultiPointCoverage']['gml:rangeSet']['gml:DataBlock']
                               ['gml:doubleOrNilReasonTupleList'].split('\n'))

    observations = []
    for idx, observation_time_data in enumerate(observation_times_list):
        observation_value_data = observation_values_list[idx]
        time_parts = observation_time_data.strip().split(' ')

        # Observation time data. Values are added next.
        observation = FMIObservation(
            datetime.utcfromtimestamp(int(time_parts[3])).replace(tzinfo=timezone.utc),
            float(time_parts[0]),
            float(time_parts[1])
        )

        # Observation variables (temperature, pressure etc..)
        for value_idx, value in enumerate(observation_value_data.strip().split(' ')):
            observation.variables[observation_types[value_idx]] = utils.float_or_none(value)

        # Sometimes the latest observation contains just NaN values. Ignore those.
        if utils.is_non_empty_observation(observation):
            observations.append(observation)

    return observations


def _get_observations_per_station(data: Dict[str, Any]) -> List[FMIStationObservation]:
    """
    This is where the magic happens. It gets all the necessary information from the FMI response combines it
    into one array of dictionaries
    :param data: Response data from FMI as xmltodict object
    :return: List of available observations from one or more weather stations
    """
    output = []

    stations = _get_stations(data)
    observations = _get_observations(data)

    # Observations are matched with stations using coordinates. [wtf.gif]
    # Since observations are ordered by coordinates, let's match them in a simple for-loop instead of trying
    # to do a pointless optimization with dictionaries ;)
    for station in stations:
        station_observations = []
        for observation in observations:
            if station.lat == observation.lat and station.lon == observation.lon:
                station_observations.append(observation)

        output.append(FMIStationObservation(station, station_observations))

    return output


def _get_closest_station_observations(lat: float,
                                      lon: float,
                                      s_observations: List[FMIStationObservation]) -> Optional[FMIStationObservation]:
    """
    Get observations from the closest station
    :param lat: Latitude
    :param lon: Longitude
    :param s_observations: Observations from all stations
    :return: Observations from the closest weather station
    """
    min_distance = sys.maxsize
    closest = None
    for station_observation in s_observations:
        station = station_observation.station
        distance = math.sqrt(((lat - float(station.lat)) ** 2) + ((lon - float(station.lon)) ** 2))
        if distance < min_distance:
            closest = station_observation
            min_distance = distance
    return closest


def _get_observations_sorted_by_distance(lat: float,
                                         lon: float,
                                         s_observations: List[FMIStationObservation]) -> List[FMIStationObservation]:
    """
    Sort observations by distance (closest first)
    :param lat: latitude
    :param lon: longitude
    :param s_observations: Observations
    :return: Observations
    """
    return sorted(s_observations,
                  key=lambda o: math.sqrt(((lat - float(o.station.lat)) ** 2) + ((lon - float(o.station.lon)) ** 2)),
                  reverse=False)



import unittest
from unittest import mock

import fmi_weather_client
import fmi_weather_client.test.test_data as test_data
from fmi_weather_client.errors import NoWeatherDataError, ServiceError


class FMIWeatherTest(unittest.TestCase):

    @mock.patch('requests.get', side_effect=test_data.mock_place_response)
    def test_get_weather_by_place(self, mock_get):
        place_weather = fmi_weather_client.weather_by_place_name('Espoo, Tapiola')
        self.assertEqual(place_weather.station_name, 'Espoo Tapiola')
        self.assertEqual(place_weather.observation_time.timestamp(), 1581331800.0)
        self.assertEqual(place_weather.temperature.value, 3.5)
        self.assertEqual(place_weather.wind_speed.value, 8.9)

    @mock.patch('requests.get', side_effect=test_data.mock_bbox_response)
    def test_get_weather_by_coordinates(self, mock_get):
        place_weather = fmi_weather_client.weather_by_coordinates(63.14343, 27.31317)
        self.assertEqual(place_weather.station_name, 'Kuopio Maaninka')
        self.assertEqual(place_weather.observation_time.timestamp(), 1581331800.0)
        self.assertEqual(place_weather.temperature.value, 2.5)
        self.assertEqual(place_weather.wind_speed.value, 5.3)

    @mock.patch('requests.get', side_effect=test_data.mock_bbox_response)
    def test_get_weather_multi_station(self, mock_get):
        place_weather = fmi_weather_client.weather_multi_station(62.99842, 27.80817)
        self.assertEqual(place_weather.station_name, 'Siilinjärvi Kuopio lentoasema')
        self.assertEqual(place_weather.observation_time.timestamp(), 1581332400.0)
        self.assertEqual(place_weather.temperature.value, 2.0)
        self.assertEqual(place_weather.humidity.value, 91.0)
        self.assertEqual(place_weather.wind_speed.value, 9.9)
        self.assertEqual(place_weather.wind_gust.value, 12.9)
        self.assertEqual(place_weather.wind_direction.value, 174.0)
        self.assertEqual(place_weather.dew_point.value, 0.8)
        self.assertEqual(place_weather.precipitation_amount, None)
        self.assertEqual(place_weather.precipitation_intensity.value, 0.0)
        self.assertEqual(place_weather.pressure.value, 963.2)
        self.assertEqual(place_weather.visibility.value, 19967.0)
        self.assertEqual(place_weather.cloud_coverage.value, 8.0)
        self.assertEqual(place_weather.snow_depth.value, 16.0)
        self.assertEqual(place_weather.wawa.value, 81.0)

    @mock.patch('requests.get', side_effect=test_data.mock_empty_response)
    def test_empty_response(self, mock_get):
        with self.assertRaises(NoWeatherDataError):
            fmi_weather_client.weather_by_coordinates(27.31317, 63.14343)

    @mock.patch('requests.get', side_effect=test_data.mock_nan_response)
    def test_nan_response(self, mock_get):
        with self.assertRaises(NoWeatherDataError):
            fmi_weather_client.weather_by_coordinates(27.31317, 63.14343)

    @mock.patch('requests.get', side_effect=test_data.mock_no_location_exception_response)
    def test_no_location_exception_response(self, mock_get):
        with self.assertRaises(NoWeatherDataError):
            fmi_weather_client.weather_by_coordinates(27.31317, 63.14343)

    @mock.patch('requests.get', side_effect=test_data.mock_other_exception_response)
    def test_other_exception_response(self, mock_get):
        with self.assertRaises(ServiceError):
            fmi_weather_client.weather_by_coordinates(27.31317, 63.14343)


if __name__ == '__main__':
    unittest.main()
